@@ -2,7 +2,7 @@
  *
  * CLEX File Manager
  *
- * Copyright (C) 2001-2006 Vlado Potisk <vlado_potisk@clex.sk>
+ * Copyright (C) 2001-2009 Vlado Potisk <vlado_potisk@clex.sk>
  *
  * CLEX is free software without warranty of any kind; see the
  * GNU General Public License as set out in the "COPYING" document
@@ -12,16 +12,16 @@
  *
  */
 
-#include <config.h>
+#include "clexheaders.h"
 
-#include <sys/types.h>	/* clex.h */
 #include <signal.h>		/* sigaction() */
 #include <unistd.h>		/* _POSIX_JOB_CONTROL */
 
-#include "clex.h"
 #include "signals.h"
 
 #include "control.h"	/* err_exit() */
+#include "inout.h"		/* curses_cbreak() */
+#include "tty.h"		/* tty_ctrlc() */
 
 static RETSIGTYPE
 int_handler(int sn)
@@ -29,6 +29,12 @@ int_handler(int sn)
 	err_exit("Signal %s caught", sn == SIGTERM ? "SIGTERM" : "SIGHUP");
 }
 
+static RETSIGTYPE
+ctrlc_handler(int unused)
+{
+	ctrlc_flag = 1;
+}
+
 void
 signal_initialize(void)
 {
@@ -41,13 +47,6 @@ signal_initialize(void)
 	sigaction(SIGINT,&act,0);
 	sigaction(SIGQUIT,&act,0);
 
-#ifdef _POSIX_JOB_CONTROL
-	/* ignore job control signals */
-	sigaction(SIGTSTP,&act,0);
-	sigaction(SIGTTIN,&act,0);
-	sigaction(SIGTTOU,&act,0);
-#endif
-
 	/* catch termination signals */
 	act.sa_handler = int_handler;
 	sigaddset(&act.sa_mask,SIGTERM);
@@ -55,3 +54,32 @@ signal_initialize(void)
 	sigaction(SIGTERM,&act,0);
 	sigaction(SIGHUP,&act,0);
 }
+
+void
+signal_ctrlc_on(void)
+{
+	struct sigaction act;
+
+	curses_cbreak();
+	tty_save();
+	tty_ctrlc();
+
+	act.sa_handler = ctrlc_handler;
+	act.sa_flags = 0;
+	sigemptyset(&act.sa_mask);
+	sigaction(SIGINT,&act,0);
+}
+
+void
+signal_ctrlc_off(void)
+{
+	struct sigaction act;
+
+	act.sa_handler = SIG_IGN;
+	act.sa_flags = 0;
+	sigemptyset(&act.sa_mask);
+	sigaction(SIGINT,&act,0);
+
+	tty_restore();
+	curses_raw();
+}
