@@ -2,7 +2,7 @@
  *
  * CLEX File Manager
  *
- * Copyright (C) 2001-2006 Vlado Potisk <vlado_potisk@clex.sk>
+ * Copyright (C) 2001-2009 Vlado Potisk <vlado_potisk@clex.sk>
  *
  * CLEX is free software without warranty of any kind; see the
  * GNU General Public License as set out in the "COPYING" document
@@ -12,38 +12,21 @@
  *
  */
 
-#include <config.h>
+#include "clexheaders.h"
 
-#include <sys/types.h>			/* clex.h */
-#include <string.h>				/* strcmp() */
+#include <stdarg.h>			/* log.h */
 
-#include "clex.h"
 #include "undo.h"
 
-#include "control.h"			/* err_exit() */
-#include "edit.h"				/* edit_update() */
-#include "inout.h"				/* win_remark() */
-#include "ustring.h"			/* USTRING */
+#include "edit.h"			/* edit_update() */
+#include "log.h"			/* msgout() */
 
 static TEXTLINE *current;
-static USTRING undo_line = { 0,0 };
+static USTRINGW undo_line = UNULL;
 static int undo_size, undo_curs, undo_offset;
 static FLAG disable = 1;
 		/* disable undo mechanism while performing undo operations */
-
-EDIT_OP this_op;	/* description of the current operation */
-
-/* initialize data structures */
-void
-undo_init(TEXTLINE *t)
-{
-	int i;
-
-	for (i = 0; i < UNDO_LEVELS; i++)
-		US_INIT(t->undo[i].save_line);
-	t->last_op.code = OP_NONE;
-	t->undo_base = t->undo_levels = t->redo_levels = 0;
-}
+static EDIT_OP this_op;	/* description of the current operation */
 
 /* clear the undo history */
 void
@@ -58,46 +41,46 @@ undo_reset(void)
 }
 
 /*
- * check if 'lstr' is the same as 'sstr' would be
+ * check if 'longstr' is the same as 'shortstr' would be
  * with 'len' chars inserted at position 'pos'
  */
 static int
-cmp_strings(const char *sstr, const char *lstr, int pos, int len)
+cmp_strings(const wchar_t *shortstr, const wchar_t *longstr, int pos, int len)
 {
 	return pos >= 0 && len >=0
-	  && (pos == 0 || strncmp(sstr,lstr,pos) == 0)
-	  && strcmp(sstr + pos,lstr + pos + len) == 0;
+	  && (pos == 0 || wcsncmp(shortstr,longstr,pos) == 0)
+	  && wcscmp(shortstr + pos,longstr + pos + len) == 0;
 }
 
 /* which edit operation was this one ? */
 static void
 tell_edit_op(void)
 {
-	int pos, len;
-	const char *before, *after;
+	int pos, diff;
+	const wchar_t *before, *after;
 
 	before = USTR(undo_line);
 	after = USTR(textline->line);
-	len = textline->size - undo_size;
-	if (len > 0) {
-		if (cmp_strings(before,after,pos = textline->curs - len,len)) {
+	diff = textline->size - undo_size;
+	if (diff > 0) {
+		if (cmp_strings(before,after,pos = textline->curs - diff,diff)) {
 			this_op.code = OP_INS;
 			this_op.pos = pos;
-			this_op.len = len;
+			this_op.len = diff;
 			return;
 		}
 	}
-	else if (len == 0) {
-		if (strcmp(before,after) == 0) {
+	else if (diff == 0) {
+		if (wcscmp(before,after) == 0) {
 			this_op.code = OP_NONE;
 			return;
 		}
 	}
 	else {
-		if (cmp_strings(after,before,pos = textline->curs,len = -len)) {
+		if (cmp_strings(after,before,pos = textline->curs,diff = -diff)) {
 			this_op.code = OP_DEL;
 			this_op.pos = pos;
-			this_op.len = len;
+			this_op.len = diff;
 			return;
 		}
 	}
@@ -115,18 +98,18 @@ undo_before(void)
 
 	disable = 0;
 	current = textline;
-	us_copy(&undo_line,USTR(textline->line));
+	usw_copy(&undo_line,USTR(textline->line));
 	undo_size   = textline->size;
 	undo_curs   = textline->curs;
 	undo_offset = textline->offset;
 }
 
-/* ... and now see what happened with it afterwards */
+/* ... and now see what happened with it after the operation */
+#define MERGE_MAX	30
 void
 undo_after(void)
 {
-	FLAG merge;
-	int idx;
+	int idx, total, delta;
 
 	if (TSET(disable) || textline == 0 || textline != current)
 		return;
@@ -135,20 +118,29 @@ undo_after(void)
 	if (this_op.code == OP_NONE)
 		return;
 
-	/* can this edit operation be merged with the previous one ? */
-	merge = (this_op.code == OP_INS && textline->last_op.code == OP_INS
-	    && this_op.len + textline->last_op.len < 10
-	    && this_op.pos == textline->last_op.pos + textline->last_op.len)
-	  || (this_op.code == OP_DEL && textline->last_op.code == OP_DEL
-	    && this_op.len == 1 && textline->last_op.len == 1
-	    && (this_op.pos == textline->last_op.pos
-	      || this_op.pos == textline->last_op.pos - 1));
-
-	textline->last_op = this_op;	/* struct copy */
 	textline->redo_levels = 0;
 
-	if (merge)
+	/*
+	 * two operations of the same type at the same position
+	 * can be merged into a single operation
+	 */
+	total = this_op.len + textline->last_op.len;
+	delta = this_op.pos - textline->last_op.pos;
+	if ((
+		this_op.code == OP_INS && textline->last_op.code == OP_INS
+		&& delta == textline->last_op.len && total < MERGE_MAX
+	  ) || (
+	  	this_op.code == OP_DEL && textline->last_op.code == OP_DEL
+		&& (delta == 0 || (delta == -1 && this_op.len == 1)) && total < MERGE_MAX
+	  )) {
+		/* merge */
+		if (this_op.code == OP_DEL)
+			textline->last_op.pos = this_op.pos;
+		textline->last_op.len = total;
 		return;
+	}
+
+	textline->last_op = this_op;	/* struct copy */
 
 	idx = (textline->undo_base + textline->undo_levels) % UNDO_LEVELS;
 	if (textline->undo_levels < UNDO_LEVELS)
@@ -156,12 +148,13 @@ undo_after(void)
 	else
 		textline->undo_base = (textline->undo_base + 1) % UNDO_LEVELS;
 
-	us_xchg(&textline->undo[idx].save_line,&undo_line);
+	usw_xchg(&textline->undo[idx].save_line,&undo_line);
 	textline->undo[idx].save_size   = undo_size;
 	textline->undo[idx].save_curs   = undo_curs;
 	textline->undo[idx].save_offset = undo_offset;
 }
 
+/* op: nonzero = undo, 0 = redo */
 static void
 undo_redo(int op)
 {
@@ -169,10 +162,9 @@ undo_redo(int op)
 
 	if (textline == 0)
 		return;
-
 	if (op) {
 		if (textline->undo_levels == 0) {
-			win_remark("undo not possible");
+			msgout(MSG_i,"undo not possible");
 			return;
 		}
 		idx = --textline->undo_levels;
@@ -180,7 +172,7 @@ undo_redo(int op)
 	}
 	else {
 		if (textline->redo_levels == 0) {
-			win_remark("redo not possible");
+			msgout(MSG_i,"redo not possible");
 			return;
 		}
 		idx = textline->undo_levels++;
@@ -188,12 +180,10 @@ undo_redo(int op)
 	}
 	idx = (textline->undo_base + idx) % UNDO_LEVELS;
 
-	us_xchg(&textline->line,&textline->undo[idx].save_line);
+	usw_xchg(&textline->line,&textline->undo[idx].save_line);
 	textline->size   = textline->undo[idx].save_size;
 	textline->curs   = textline->undo[idx].save_curs;
 	textline->offset = textline->undo[idx].save_offset;
-
-	us_xchg(&textline->undo[idx].save_line,&undo_line);
 	textline->undo[idx].save_size   = undo_size;
 	textline->undo[idx].save_curs   = undo_curs;
 	textline->undo[idx].save_offset = undo_offset;
