@@ -2,7 +2,7 @@
  *
  * CLEX File Manager
  *
- * Copyright (C) 2001-2006 Vlado Potisk <vlado_potisk@clex.sk>
+ * Copyright (C) 2001-2009 Vlado Potisk <vlado_potisk@clex.sk>
  *
  * CLEX is free software without warranty of any kind; see the
  * GNU General Public License as set out in the "COPYING" document
@@ -12,34 +12,29 @@
  *
  */
 
-#include <config.h>
+#include "clexheaders.h"
 
-#include <sys/types.h>			/* clex.h */
-#include <errno.h>				/* errno */
 #include <stdarg.h>				/* va_list */
 #include <stdlib.h>				/* free() */
 #include <string.h>				/* strlen() */
-#include <unistd.h>				/* getcwd() */
-
-#include "clex.h"
-#include "ustring.h"
 
 #include "util.h"				/* emalloc() */
 
-extern int errno;
-
 /*
- * The USTRING structure (defined in clex.h) can store a string
+ * The USTRING structure (defined in ustring.h) can store a string
  * of unlimited length. The memory is allocated dynamically.
  *
  * - to initialize (to NULL ptr value) before first use:
- *     static USTRING us = { 0, 0 };
- *       or
+ *   - static and global variables are initialized by default,
+ *     but if you prefer explicit initialization:
+ *       static USTRING us = UNULL;
+ *         which is equivalent to
+ *       static USTRING us = { 0, 0 };
+ *   - dynamic variables are initialized this way:
  *     US_INIT(ustring);
  * - to re-initialize, e.g. before deallocating dynamic USTRING:
  *     us_reset();
- * - to store a string
- *   (NULL ptr cannot be stored, but see us_copy() function below):
+ * - to store a string: (us_copy accepts also a null ptr)
  *     us_copy();
  *       or
  *     us_copyn();
@@ -55,7 +50,7 @@ extern int errno;
  */
 
 /* these are tunable parameters */
-#define ALLOC_UNIT			24	/* in bytes */
+/* ALLOC_UNIT in ustring.h */
 #define MINIMUM_FREE		(4 * ALLOC_UNIT)
 
 /*
@@ -71,6 +66,7 @@ extern int errno;
 	((1 + (RQ - 1) / ALLOC_UNIT) * ALLOC_UNIT)
 /* note that ROUND_ALLOC(0) is ALLOC_UNIT and not 0 */
 
+/* clear the data and free the memory */
 void
 us_reset(USTRING *pustr)
 {
@@ -81,26 +77,68 @@ us_reset(USTRING *pustr)
 	pustr->USstr = 0;
 }
 
-/* us_setsize() makes room for 'memreq' characters */
+/* wchar version */
 void
-us_setsize(USTRING *pustr, size_t memreq)
+usw_reset(USTRINGW *pustr)
 {
-	if (SHOULD_CHANGE_ALLOC(memreq)) {
+	if (pustr->USalloc) {
+		free(pustr->USstr);
+		pustr->USalloc = 0;
+	}
+	pustr->USstr = 0;
+}
+
+/* us_setsize() makes room for at least 'req' characters */
+size_t
+us_setsize(USTRING *pustr, size_t req)
+{
+	if (SHOULD_CHANGE_ALLOC(req)) {
 		if (pustr->USalloc)
 			free(pustr->USstr);
-		pustr->USalloc = ROUND_ALLOC(memreq);
+		pustr->USalloc = ROUND_ALLOC(req);
 		pustr->USstr = emalloc(pustr->USalloc);
 	}
+
+	return pustr->USalloc;	/* real buffer size is returned */
+}
+
+/* wchar version; note that 'req' is in characters, not bytes */
+size_t
+usw_setsize(USTRINGW *pustr, size_t req)
+{
+	if (SHOULD_CHANGE_ALLOC(req)) {
+		if (pustr->USalloc)
+			free(pustr->USstr);
+		pustr->USalloc = ROUND_ALLOC(req);
+		pustr->USstr = emalloc(sizeof(wchar_t) * pustr->USalloc);
+	}
+
+	return pustr->USalloc;
 }
 
 /* like us_setsize(), but preserving contents */
-void
-us_resize(USTRING *pustr, size_t memreq)
+size_t
+us_resize(USTRING *pustr, size_t req)
 {
-	if (SHOULD_CHANGE_ALLOC(memreq)) {
-		pustr->USalloc = ROUND_ALLOC(memreq);
+	if (SHOULD_CHANGE_ALLOC(req)) {
+		pustr->USalloc = ROUND_ALLOC(req);
 		pustr->USstr = erealloc(pustr->USstr,pustr->USalloc);
 	}
+
+	return pustr->USalloc;
+}
+
+
+/* wchar version */
+size_t
+usw_resize(USTRINGW *pustr, size_t req)
+{
+	if (SHOULD_CHANGE_ALLOC(req)) {
+		pustr->USalloc = ROUND_ALLOC(req);
+		pustr->USstr = erealloc(pustr->USstr,sizeof(wchar_t) * pustr->USalloc);
+	}
+
+	return pustr->USalloc;
 }
 
 /* quick alternative to copy */
@@ -119,22 +157,49 @@ us_xchg(USTRING *s1, USTRING *s2)
 	s2->USalloc = xalloc;
 }
 
+/* wchar version */
 void
+usw_xchg(USTRINGW *s1, USTRINGW *s2)
+{
+	wchar_t *xstr;
+	size_t xalloc;
+
+	xstr      = s1->USstr;
+	s1->USstr = s2->USstr;
+	s2->USstr = xstr;
+
+	xalloc      = s1->USalloc;
+	s1->USalloc = s2->USalloc;
+	s2->USalloc = xalloc;
+}
+
+char *
 us_copy(USTRING *pustr, const char *src)
 {
-#if 0
-/* uncomment this code to enable NULL ptr assignment */
 	if (src == 0) {
 		us_reset(pustr);
-		return;
+		return 0;
 	}
-#endif
 	us_setsize(pustr,strlen(src) + 1);
 	strcpy(pustr->USstr,src);
+	return pustr->USstr;
+}
+
+/* wchar version */
+wchar_t *
+usw_copy(USTRINGW *pustr, const wchar_t *src)
+{
+	if (src == 0) {
+		usw_reset(pustr);
+		return 0;
+	}
+	usw_setsize(pustr,wcslen(src) + 1);
+	wcscpy(pustr->USstr,src);
+	return pustr->USstr;
 }
 
 /* note: us_copyn() adds terminating null byte */
-void
+char *
 us_copyn(USTRING *pustr, const char *src, size_t len)
 {
 	char *dst;
@@ -144,6 +209,21 @@ us_copyn(USTRING *pustr, const char *src, size_t len)
 	dst[len] = '\0';
 	while (len-- > 0)
 		dst[len] = src[len];
+	return dst;
+}
+
+/* wchar version */
+wchar_t *
+usw_copyn(USTRINGW *pustr, const wchar_t *src, size_t len)
+{
+	wchar_t *dst;
+
+	usw_setsize(pustr,len + 1);
+	dst = pustr->USstr;
+	dst[len] = L'\0';
+	while (len-- > 0)
+		dst[len] = src[len];
+	return dst;
 }
 
 /* concatenation: us_cat(&ustring, str1, str2, ..., strN, (char *)0); */
@@ -168,41 +248,24 @@ us_cat(USTRING *pustr, ...)
 	va_end(argptr);
 }
 
-/* USTRING version of getcwd() */
-int
-get_cwd_us(USTRING *pustr)
+/* wchar version */
+void
+usw_cat(USTRINGW *pustr, ...)
 {
-	us_setsize(pustr,ALLOC_UNIT);
-	for (;/* until return*/;) {
-		if (getcwd(pustr->USstr,pustr->USalloc))
-			return 0;
-		if (errno != ERANGE)
-			return -1;
-		/* increase buffer */
-		us_setsize(pustr,pustr->USalloc + ALLOC_UNIT);
-	}
-}
+	size_t len;
+	wchar_t *str;
+	va_list argptr;
 
-/* USTRING version of readlink() */
-int
-get_link_us(USTRING *pustr, const char *path)
-{
-#ifndef HAVE_READLINK
-	return -1;
-#else
-	int len;
+	va_start(argptr,pustr);
+	for (len = 1; (str = va_arg(argptr, wchar_t *)); )
+		len += wcslen(str);
+	va_end(argptr);
+	usw_setsize(pustr,len);
 
-	us_setsize(pustr,ALLOC_UNIT);
-	for (;/* until return*/;) {
-		len = readlink(path,pustr->USstr,pustr->USalloc);
-		if (len == -1)
-			return -1;
-		if (len < pustr->USalloc) {
-			pustr->USstr[len] = '\0';
-			return 0;
-		}
-		/* increase buffer */
-		us_setsize(pustr,pustr->USalloc + ALLOC_UNIT);
+	va_start(argptr,pustr);
+	for (len = 0; (str = va_arg(argptr, wchar_t *)); ) {
+		wcscpy(pustr->USstr + len,str);
+		len += wcslen(str);
 	}
-#endif
+	va_end(argptr);
 }
