@@ -2,7 +2,7 @@
  *
  * CLEX File Manager
  *
- * Copyright (C) 2001-2006 Vlado Potisk <vlado_potisk@clex.sk>
+ * Copyright (C) 2001-2009 Vlado Potisk <vlado_potisk@clex.sk>
  *
  * CLEX is free software without warranty of any kind; see the
  * GNU General Public License as set out in the "COPYING" document
@@ -12,33 +12,30 @@
  *
  */
 
-#include <config.h>
+#include "clexheaders.h"
 
-#include <sys/types.h>	/* clex.h */
-#include <ctype.h>		/* iscntrl() */
-#include <signal.h>		/* signal() */
+#include <sys/wait.h>	/* waitpid() */
+#include <sys/stat.h>	/* open() */
+#include <ctype.h>		/* isprint() */
+#include <fcntl.h>		/* open() */
+#include <signal.h>		/* sigaction() */
 #include <stdarg.h>		/* va_list */
-#include <stdio.h>		/* printf() */
+#include <stdio.h>		/* fputs() */
 #include <stdlib.h>		/* getenv() */
 #include <string.h>		/* strchr() */
 #include <unistd.h>		/* fork() */
 
-/* waitpid() */
-#ifdef HAVE_SYS_WAIT_H
-# include <sys/wait.h>
-#endif
-
-#include "clex.h"
 #include "xterm_title.h"
 
-#include "cfg.h"		/* config_num() */
+#include "cfg.h"		/* cfg_num() */
+#include "log.h"		/* msgout() */
 #include "util.h"		/* estrdup() */
 
 static FLAG enabled = 0;
-static const char *oldtitle = 0;
+static const char *old_title = 0, default_title[] = "terminal";
 
-#define XPROP_TIMEOUT	6	/* timeout for xprop command in seconds */
-#define CMD_STR			20	/* name of executed command in the title */
+#define XPROP_TIMEOUT	4	/* timeout for the xprop command in seconds */
+#define CMD_STR			48	/* buffer size for the executed command's name in the title */
 
 void
 xterm_title_initialize(void)
@@ -48,19 +45,16 @@ xterm_title_initialize(void)
 }
 
 /*
- * run the command
- *   xprop -id $WINDOWID WM_NAME
+ * run the command: xprop -id $WINDOWID WM_NAME 2>/dev/null
  * to get the current xterm title
- *
- * errors are not reported, we don't want to confuse the user
  */
-static const char *
+static char *
 get_title(void)
 {
-	int fd[2];
+	int fd[2], efd;
 	ssize_t rd;
+	char *p1, *p2, title[192];
 	const char *wid;
-	char *p1, *p2, title[128];
 	pid_t pid;
 	struct sigaction act;
 
@@ -72,6 +66,8 @@ get_title(void)
 
 	if (pid == 0) {
 		/* this is the child process */
+		logfile_close();
+
 		close(fd[0]);	/* close read end */
 		if (fd[1] != STDOUT_FILENO) {
 			if (dup2(fd[1], STDOUT_FILENO) != STDOUT_FILENO)
@@ -79,6 +75,13 @@ get_title(void)
 			close(fd[1]);
 		}
 
+		efd = open("/dev/null",O_RDONLY);
+		if (efd != STDERR_FILENO) {
+			if (dup2(efd, STDERR_FILENO) != STDERR_FILENO)
+				_exit(126);
+			close(efd);
+		}
+
 		act.sa_handler = SIG_DFL;
 		act.sa_flags = 0;
 		sigemptyset(&act.sa_mask);
@@ -90,7 +93,7 @@ get_title(void)
 	}
 
 	/* parent continues here */
-	/* do not return before waitpid() or you create a zombie */
+	/* do not return before waitpid() - otherwise you'll create a zombie */
 	close(fd[1]);	/* close write end */
 	rd = read_fd(fd[0],title,sizeof(title) - 1);
 	close(fd[0]);
@@ -117,25 +120,23 @@ get_title(void)
 void
 xterm_title_reconfig(void)
 {
-	const char *term;
-
-	enabled = 0;
-	/* 0 = disabled */
-	if (config_num(CFG_XTERM_TITLE) == 0)
+	enabled = cfg_num(CFG_XTERM_TITLE);
+	if (!enabled || (old_title != 0 && old_title != default_title))
 		return;
 
-	/* 1 = automatic -> check $TERM for xterm, kterm or dtterm */
-	if (config_num(CFG_XTERM_TITLE) == 1) {
-		if ( (term = getenv("TERM")) == 0)
-			return;
-		if (strncmp(term,"xterm",5) && strncmp(term,"kterm",5)
-		  && strncmp(term,"dtterm",6))
-			return;
+	/* the terminal must be "xterm" or compatible or at least it must be an X application */
+	if (!disp_data.xterm && !disp_data.xwin) {
+		msgout(MSG_NOTICE,"Disabling the terminal title change feature (no xterm and no X Window)");
+		enabled = 0;
+		return;
+	}
+
+	if ((old_title = get_title()) == 0) {
+		msgout(MSG_NOTICE,"Could not get the current terminal window title"
+		  " because the command \"xprop -id $WINDOWID WM_NAME\" has failed."
+		  " CLEX will not be able to restore the original title when it terminates");
+		old_title = default_title;
 	}
-	/* otherwise 2 = enabled */
-	if (oldtitle == 0 && (oldtitle = get_title()) == 0)
-		oldtitle = "terminal";
-	enabled = 1;
 }
 
 static void
@@ -149,7 +150,8 @@ set_xtitle(const char *str1, ...)
 	fputs(str1,stdout);
 	va_start(argptr,str1);
 	while ( (strn = va_arg(argptr, char *)) )
-		fputs(strn,stdout);
+		if (*strn)
+			fputs(strn,stdout);
 	va_end(argptr);
 
 	/* Xterm FAQ claims \007 is incorrect, but everybody uses it */
@@ -160,8 +162,8 @@ set_xtitle(const char *str1, ...)
 void
 xterm_title_set(int busy, const char *cmd)
 {
-	char short_cmd[CMD_STR];
-	FLAG longcmd;
+	char title_cmd[CMD_STR];
+	FLAG islong;
 	int i, ch;
 
 	if (!enabled)
@@ -169,27 +171,27 @@ xterm_title_set(int busy, const char *cmd)
 
 	if (cmd == 0) {
 		/* CLEX is idle */
-		set_xtitle("clex: ", clex_data.login_at_host, (char *)0);
+		set_xtitle("clex: ", user_data.login,"@",user_data.host, (char *)0);
 		return;
 	}
 
-	/* CLEX is executing or was executing 'cmd' */
-	for (longcmd = 0, i = 0; (ch = (unsigned char)cmd[i]); i++) {
+	/* CLEX is executing (busy is true) or was executing (busy is false) 'cmd' */
+	for (islong = 0, i = 0; (ch = (unsigned char)cmd[i]); i++) {
 		if (i == CMD_STR - 1) {
-			longcmd = 1;
+			islong = 1;
 			break;
 		}
-		short_cmd[i] = iscntrl(ch) ? '?' : ch;
+		title_cmd[i] = isprint(ch) ? ch : '?';
 	}
-	short_cmd[i] = '\0';
-	set_xtitle("clex: ", busy ? "" : "(", short_cmd,
-	  longcmd ? "..." : "", busy ? "" : ")", (char *)0);
+	title_cmd[i] = '\0';
+	set_xtitle(busy ? "" : "[", "clex: ", title_cmd, islong ? "..." : "",
+	  busy ? (char *)0 : "]", (char *)0);
 }
 
-/* note: this is a cleanup function */
+/* this is a cleanup function (see err_exit() in control.c) */
 void
 xterm_title_restore(void)
 {
 	if (enabled)
-		set_xtitle(oldtitle,(char *)0);
+		set_xtitle(old_title,(char *)0);
 }
